use std::collections::{HashMap, HashSet};
use std::env::args_os;
use std::ffi::{OsStr, OsString};
use std::ops::RangeBounds;

/// A command line argument parser.
#[derive(Debug, Clone)]
pub struct Parser {
    args: Vec<Arg>,

    // Used to ensure we don't get duplicate arguments.
    id_set: HashSet<String>,
    long_set: HashSet<String>,
    short_set: HashSet<String>,
}

impl Parser {
    pub fn new() -> Parser {
        Parser {
            args: Vec::new(),
            id_set: HashSet::new(),
            long_set: HashSet::new(),
            short_set: HashSet::new(),
        }
    }

    /// Add a flag (bool) argument.
    ///
    /// - `id`: the argument identifier, used for fetching argument
    ///   matches.
    /// - `flags`: the long and/or short argument flag strings.  Must be
    ///   in the form "-f" or "--flag".  You can pass as many as you
    ///   like, all of which will be considered equivalent during
    ///   parsing.  But there must be at least one.
    /// - `doc`: the documentation string to use in the generated help.
    ///   Pass an empty string to indicate no documentation.
    pub fn add_flag(&mut self, id: &str, flags: &[&str], doc: &str) {
        let (long_flags, short_flags) = self.validate_and_process_arg(id, flags);

        self.args.push(Arg {
            arg_type: ArgType::Flag,
            id: id.into(),
            value_label: String::new(),
            long_flags: long_flags,
            short_flags: short_flags,
            acceptable_count: (0, None),
            doc: doc.into(),
        });
    }

    /// Add a standard argument, that takes a value.
    pub fn add_argument(
        &mut self,
        id: &str,
        flags: &[&str],
        doc: &str,
        value_label: &str,
        required: bool,
    ) {
        let (long_flags, short_flags) = self.validate_and_process_arg(id, flags);

        self.args.push(Arg {
            arg_type: ArgType::Arg,
            id: id.into(),
            value_label: value_label.into(),
            long_flags: long_flags,
            short_flags: short_flags,
            acceptable_count: (if required { 1 } else { 0 }, None),
            doc: doc.into(),
        });
    }

    /// Add a positional argument.
    ///
    /// Unlike flags and standard arguments, positional arguments are
    /// parsed in the order they're added.  Because of their nature,
    /// they have some additional considerations:
    ///
    /// - All required positional arguments must precede all optional
    ///   positional arguments.
    /// - There can at most be a single positional multi-argument,
    ///   which must come last.  (See `add_positional_multi_argument()`.)
    pub fn add_positional_argument(
        &mut self,
        id: &str,
        doc: &str,
        value_label: &str,
        required: bool,
    ) {
        let (_, _) = self.validate_and_process_arg(id, &[]);

        self.args.push(Arg {
            arg_type: ArgType::PosArg,
            id: id.into(),
            value_label: value_label.into(),
            long_flags: Vec::new(),
            short_flags: Vec::new(),
            acceptable_count: (if required { 1 } else { 0 }, Some(1)),
            doc: doc.into(),
        });
    }

    pub fn add_positional_multi_argument(
        &mut self,
        id: &str,
        doc: &str,
        value_label: &str,
        required: bool,
    ) {
        let (_, _) = self.validate_and_process_arg(id, &[]);

        self.args.push(Arg {
            arg_type: ArgType::PosArg,
            id: id.into(),
            value_label: value_label.into(),
            long_flags: Vec::new(),
            short_flags: Vec::new(),
            acceptable_count: (if required { 1 } else { 0 }, None),
            doc: doc.into(),
        });
    }

    //----------------

    pub fn parse(self) -> ParsedArguments {
        todo!()
    }

    //----------------

    /// Returns (long, short) pair, each of which is a Vec of argument strings with
    /// the leading hyphens stripped off.
    fn validate_and_process_arg(&mut self, id: &str, flags: &[&str]) -> (Vec<String>, Vec<String>) {
        if self.id_set.contains(id) {
            panic!(
                "Error: attempted to add argument with a duplicate ID \"{}\".",
                id
            );
        }
        self.id_set.insert(id.into());

        let mut long_flags = Vec::new();
        let mut short_flags = Vec::new();

        for &flag in flags {
            // Ensure no whitespace.
            if flag.len() != flag.trim().len() || flag.split_whitespace().count() > 1 {
                panic!(
                    "Error: attempted to add argument \"{}\" which contains whitespace.",
                    flag
                );
            }
            // Long flags.
            else if flag.starts_with("--") && flag.len() > 2 {
                if self.long_set.contains(flag) {
                    panic!(
                        "Error: attempted to add duplicate long argument \"{}\".",
                        flag
                    );
                }
                self.long_set.insert(flag.into());
                long_flags.push((&flag[2..]).into());
            }
            // Check if it's a valid short flag (should only have one character
            // after the hyphen).
            else if flag.starts_with("-") && flag.chars().count() == 2 {
                if self.short_set.contains(flag) {
                    panic!(
                        "Error: attempted to add duplicate short argument \"{}\".",
                        flag
                    );
                }
                self.short_set.insert(flag.into());
                short_flags.push((&flag[1..]).into());
            }
            // Not a valid flag.
            else {
                panic!(
                    "Error: attempted to add argument \"{}\", which isn't a valid argument string.",
                    flag
                )
            }
        }

        (long_flags, short_flags)
    }
}

/// Parsed command line arguments.
#[derive(Debug, Clone)]
pub struct ParsedArguments {
    // A list of all passed arguments, in the order they were
    // passed.  Each item is a `(argument_id, argument_value)` tuple.
    // Boolean flags do not have a value.
    arguments: Vec<(String, Option<OsString>)>,

    // Maps from an argument's ID to the indices in `arguments` where it
    // is stored.
    id_map: HashMap<String, Vec<usize>>, // Argument ID -> index list
}

//-------------------------------------------------------------

#[derive(Debug, Copy, Clone, Eq, PartialEq)]
enum ArgType {
    Flag,
    Arg,
    PosArg,
}

/// Argument specification.
#[derive(Debug, Clone)]
struct Arg {
    arg_type: ArgType,
    id: String,
    value_label: String,

    // Long and short versions of the argument flag.  E.g. "--curve" and
    // "-c", but without the leading dashes.
    long_flags: Vec<String>,
    short_flags: Vec<String>,

    // How many instances of the argument can be present, specified
    // as a range.
    //
    // For example:
    // - (0, None): An argument that can show up any number of times,
    //              including not at all.
    // - (0, 1):    An argument that can either be absent or show up
    //              precisely once.
    // - (1, 1):    An argument that must show up precisely once.
    // - (1, None): An argument that must show up at least once.
    // - (2, 9):    An argument that must show up at least twice, but
    //              no more than 9 times.
    acceptable_count: (usize, Option<usize>),

    // Documentation string, for generated help.
    doc: String,
}
